# coding: utf-8
"""
class Timeseries (Base)
=======================
"""
from sqlalchemy import Column, DateTime, Float, ForeignKey, Integer, text, String, \
                       Text, CheckConstraint, UniqueConstraint, PrimaryKeyConstraint, \
                       Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.dialects.postgresql import JSONB
from toardb.stationmeta.models import StationmetaCore
from toardb.variables.models import Variable
from .models_programme import TimeseriesProgramme
from toardb.base import Base


TIMESERIES_ID_SEQ = Sequence('timeseries_id_seq')  # define sequence explicitly
class Timeseries(Base):
    """ Table "public.timeseries"

    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    |       Column        |           Type           | Collation | Nullable |                Default                 |
    +=====================+==========================+===========+==========+========================================+
    | id                  | integer                  |           | not null | nextval('timeseries_id_seq'::regclass) |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | label               | character varying(128)   |           | not null |                                        |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | order               | integer                  |           | not null |                                        |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | access_rights       | integer                  |           | not null |                                        |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | sampling_frequency  | integer                  |           | not null |                                        |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | aggregation         | integer                  |           | not null |                                        |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | source              | integer                  |           | not null | 1                                      |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | data_start_date     | timestamp with time zone |           | not null |                                        |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | data_end_date       | timestamp with time zone |           | not null |                                        |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | measurement_method  | integer                  |           | not null | 1                                      |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | sampling_height     | double precision         |           | not null |                                        |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | additional_metadata | jsonb                    |           | not null |                                        |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | date_added          | timestamp with time zone |           | not null | now()                                  |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | date_modified       | timestamp with time zone |           | not null | now()                                  |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | station_id          | integer                  |           |          |                                        |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | variable_id         | integer                  |           |          |                                        |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    | programme_id        | integer                  |           | not null | 0                                      |
    +---------------------+--------------------------+-----------+----------+----------------------------------------+
    Indexes:
        "timeseries_pkey" PRIMARY KEY, btree (id)
        "timeseries_station_id_variable_id_label_uniq" UNIQUE CONSTRAINT, btree (station_id, variable_id, label)
        "timeseries_station_id" btree (station_id)
        "timeseries_variable_id" btree (variable_id)
    Check constraints:
        "timeseries_access_rights_check" CHECK (access_rights >= 0)
        "timeseries_aggregation_check" CHECK (aggregation >= 0)
        "timeseries_order_check" CHECK ("order" >= 0)
        "timeseries_sampling_frequency_check" CHECK (sampling_frequency >= 0)
    Foreign-key constraints:
        "timeseries_station_id_fk_stationmeta_core_id" FOREIGN KEY (station_id) REFERENCES stationmeta_core(id) DEFERRABLE INITIALLY DEFERRED
        "timeseries_variable_id_fk_variables_id" FOREIGN KEY (variable_id) REFERENCES variables(id) DEFERRABLE INITIALLY DEFERREDForeign-key constraints:
        "timeseries_access_rights_fk_da_vocabulary_enum_val" FOREIGN KEY (access_rights) REFERENCES da_vocabulary(enum_val)
        "timeseries_aggregation_fk_at_vocabulary_enum_val" FOREIGN KEY (aggregation) REFERENCES at_vocabulary(enum_val)
        "timeseries_measurement_method_fk_mm_vocabulary_enum_val" FOREIGN KEY (measurement_method) REFERENCES mm_vocabulary(enum_val)
        "timeseries_sampling_frequency_fk_sf_vocabulary_enum_val" FOREIGN KEY (sampling_frequency) REFERENCES sf_vocabulary(enum_val)
        "timeseries_source_fk_ds_vocabulary_enum_val" FOREIGN KEY (source) REFERENCES ds_vocabulary(enum_val)
        "timeseries_programme_id_fk_timeseries_programmes_id" FOREIGN KEY (programme_id) REFERENCES timeseries_programmes(id)
    Referenced by:
        TABLE "data" CONSTRAINT "data_timeseries_id_fk_timeseries_id" FOREIGN KEY (timeseries_id) REFERENCES timeseries(id) DEFERRABLE INITIALLY DEFERRED
        TABLE "timeseries_timeseries_annotations" CONSTRAINT "timeseries_timeseries_annotations_timeseries_id_fk_timeseries_id" FOREIGN KEY (timeseries_id) REFERENCES timeseries(id) DEFERRABLE INITIALLY DEFERRED
        TABLE "timeseries_timeseries_roles" CONSTRAINT "timeseries_timeseries_roles_timeseries_id_fk_timeseries_id" FOREIGN KEY (timeseries_id) REFERENCES timeseries(id) DEFERRABLE INITIALLY DEFERRED
        TABLE "timeseries_timeseries_programmes" CONSTRAINT "timeseries_timeseries_programmes_timeseries_id_fk_timeseries_id" FOREIGN KEY (timeseries_id) REFERENCES timeseries(id) DEFERRABLE INITIALLY DEFERRED
    """

    __tablename__ = 'timeseries'
    __table_args__ = (
        CheckConstraint('"order" >= 0'),
        CheckConstraint('access_rights >= 0'),
        CheckConstraint('aggregation >= 0'),
        CheckConstraint('sampling_frequency >= 0'),
        UniqueConstraint('station_id', 'variable_id', 'label'),
    )

    id = Column(Integer, TIMESERIES_ID_SEQ, primary_key=True, server_default=TIMESERIES_ID_SEQ.next_value())
    label = Column(String(128), nullable=False)
    order = Column(Integer, nullable=False)
    access_rights = Column(ForeignKey('da_vocabulary.enum_val'), nullable=False)
    sampling_frequency = Column(ForeignKey('sf_vocabulary.enum_val'), nullable=False)
    aggregation = Column(ForeignKey('at_vocabulary.enum_val'), nullable=False)
    source = Column(ForeignKey('ds_vocabulary.enum_val'), nullable=False, server_default=text("1"))
    data_start_date = Column(DateTime(True), nullable=False)
    data_end_date = Column(DateTime(True), nullable=False)
    measurement_method = Column(ForeignKey('mm_vocabulary.enum_val'), nullable=False, server_default=text("1"))
    sampling_height = Column(Float(53), nullable=False)
    additional_metadata = Column(JSONB(astext_type=Text()), nullable=True)
    date_added = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))
# do not use string declaration here (not working for pytest)
# use the explicit class name here,
# see: https://groups.google.com/forum/#!topic/sqlalchemy/YjGhE4d6K4U
    station_id = Column(ForeignKey(StationmetaCore.id, deferrable=True, initially='DEFERRED'), index=True)
    variable_id = Column(ForeignKey(Variable.id, deferrable=True, initially='DEFERRED'), index=True)
    programme_id = Column(ForeignKey(TimeseriesProgramme.id), nullable=False, server_default=text("0"))

    # for the nested view
    # problems with station and coordinates in nested views!!!
    station = relationship('StationmetaCore_WithoutCoords')
    variable = relationship('Variable')
    programme = relationship('TimeseriesProgramme')

#   da_vocabulary = relationship('DaVocabulary')
#   at_vocabulary = relationship('AtVocabulary')
#   mm_vocabulary = relationship('MmVocabulary')
#   sf_vocabulary = relationship('SfVocabulary')
#   ds_vocabulary = relationship('DsVocabulary')

